
#if !defined(__NativeInterpreterVmCallbacks_h__)
#define __NativeInterpreterVmCallbacks_h__


#include "NativeInterpreterTypes.h"
#include "NativeInterpreterMacros.h"
#include "ExternalCall.h"
#include "MemoryAccess.h"

#include <stdint.h>
#include <limits.h>

namespace nanojit
{
namespace interpreter
{

bool doTest(IntepreterContext* context, unsigned int ra, unsigned int rb, LOpcode op, LTy type);


template<typename T>
const uint8_t* readImm(T& value, const uint8_t* stream);

template<>
inline const uint8_t* readImm<uint32_t>(uint32_t& value, const uint8_t* stream) {
	READ_IMMI32(value, stream);
	return stream;
}

#if defined(NANOJIT_64BIT)
template<>
inline const uint8_t* readImm<uint64_t>(uint64_t& value, const uint8_t* stream) {
	READ_IMMI64(value, stream);
	return stream;
}
#endif

template<>
inline const uint8_t* readImm<float32_t>(float32_t& value, const uint8_t* stream) {
	READ_IMMI32(value.v, stream);
	return stream;
}

template<>
inline const uint8_t* readImm<float64_t>(float64_t& value, const uint8_t* stream) {
	READ_IMMI64(value.v, stream);
	return stream;
}

template<typename T>
inline const uint8_t* callBackImm(IntepreterContext* context, const uint8_t* stream) {
	uint32_t r;
	T v;
	READ_REGISTER_INDEX(r, stream);
	stream = readImm<T>(v, stream);
	context->set<T>(r, v);
	return stream;
}

template<typename T>
inline const uint8_t* callbackMov(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;
 	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);
	context->set<T>(rd, context->get<T>(rs));
}


template<typename T>
inline const uint8_t* callbackAdd(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	T res	= context->get<T>(r0) + context->get<T>(r1);
	context->set<T>(rd, res);

	return stream;
}

template<typename T>
inline const uint8_t* callbackSub(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	T res	= context->get<T>(r0) - context->get<T>(r1);
	context->set<T>(rd, res);

	return stream;
}

template<typename T>
inline const uint8_t* callbackAddc(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	T carryValue	= context->getFlag(ArithFlag_Carry) ? 1 : 0;
	T res			= context->get<T>(r0) + context->get<T>(r1) + carryValue;
	context->set<T>(rd, res);

	return stream;
}

template<typename T>
inline const uint8_t* callbackSubb(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	T borrowValue	= context->getFlag(ArithFlag_Borrow) ? 1 : 0;
	T res			= context->get<T>(r0) - context->get<T>(r1) - borrowValue;
	context->set<T>(rd, res);

	return stream;
}

template<typename T>
inline const uint8_t* callbackNot(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	T res	= ~context->get<T>(rs);
	context->set<T>(rd, res);

	return stream;
}

template<typename T>
inline const uint8_t* callbackNeg(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	T res	= -context->get<T>(rs);
	context->set<T>(rd, res);

	return stream;
}

template<typename T>
inline const uint8_t* callbackOr(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	T res	= context->get<T>(r0) | context->get<T>(r1);
	context->set<T>(rd, res);

	return stream;
}

template<typename T>
inline const uint8_t* callbackXor(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	T res	= context->get<T>(r0) ^ context->get<T>(r1);
	context->set<T>(rd, res);

	return stream;
}

template<typename T>
inline const uint8_t* callbackAnd(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	T res	= context->get<T>(r0) & context->get<T>(r1);
	context->set<T>(rd, res);

	return stream;
}

template<typename T>
inline const uint8_t* callbackAndc(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	T res	= context->get<T>(r0) & ~context->get<T>(r1);
	context->set<T>(rd, res);

	return stream;
}


template<typename T>
inline const uint8_t* callbackLsh(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	T res	= context->get<T>(r0) << context->get<uint32_t>(r1);
	context->set<T>(rd, res);

	return stream;
}

template<typename T>
inline const uint8_t* callbackRsh(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	T res	= context->get<T>(r0) >> context->get<uint32_t>(r1);
	context->set<T>(rd, res);

	return stream;
}


template<typename T>
inline const uint8_t* callbackRor(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	uint32_t	count		= context->get<uint32_t>(r1);
	T			regValue	= context->get<T>(r0);
	T			res			= regValue;

	if (count != 0) {
		uint32_t	invCount	= (sizeof(T) * CHAR_BIT) - count;
		T			res0		= regValue >> count;
		T			res1		= regValue << invCount;

		res			= res0 | res1;
	}

	context->set<T>(rd, res);
	return stream;
}

template<typename T>
inline const uint8_t* callbackRcr(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	uint32_t	count		= context->get<uint32_t>(r1); 
	uint32_t	invCount	= (sizeof(T) * CHAR_BIT) - count;
	T			regValue	= context->get<T>(r0);
	T			oldCarry	= context->getFlag(ArithFlag_Carry) ? 1 : 0;
	T			res0		= regValue >> count;
	T			res1		= invCount < 0x1F ? (regValue << (invCount + 1)) : 0;
	T			resCarry	= oldCarry << invCount;
	
	T			res			= res0 | res1 | resCarry;

	context->set<T>(rd, res);
	return stream;
}


template<typename T>
inline const uint8_t* callbackAbs(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	T			regValue	= context->get<T>(rs);
	T			res			= abs(regValue);

	context->set<T>(rd, res);
	return stream;
}

template<typename T>
inline const uint8_t* callbackSqrt(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	T			regValue	= context->get<T>(rs);
	T			res			= sqrt(regValue);

	context->set<T>(rd, res);
	return stream;
}

template<typename T>
inline const uint8_t* callbackDiv(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);

	T			regValue0	= context->get<T>(r0);
	T			regValue1	= context->get<T>(r1);
	T			res			= regValue0 / regValue1;

	context->set<T>(rd, res);
	return stream;
}


template<typename T>
inline const uint8_t* callbackRecip(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	T			regValue	= context->get<T>(rs);
	T			res			= rcp(regValue);

	context->set<T>(rd, res);
	return stream;
}
template<typename T>
inline const uint8_t* callbackRsqrt(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	T			regValue	= context->get<T>(rs);
	T			res			= rsqrt(regValue);

	context->set<T>(rd, res);
	return stream;
}



template<typename T>
inline const uint8_t* callbackMul(IntepreterContext* context, const uint8_t* stream) {

	uint32_t rd, r0, r1;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(r0, stream);
	READ_REGISTER_INDEX(r1, stream);
	
	T	v0			= context->get<T>(r0);
	T	v1			= context->get<T>(r1);

	T	vd			= v0 * v1;

	context->set<T>(rd, vd);

	return stream;
}

template<typename T, typename TL>
inline const uint8_t* callbackMulLong(IntepreterContext* context, const uint8_t* stream) {

	uint32_t rdlo, rdhi, ra, rb;

	READ_REGISTER_INDEX(rdlo, stream);
	READ_REGISTER_INDEX(rdhi, stream);
	READ_REGISTER_INDEX(ra, stream);
	READ_REGISTER_INDEX(rb, stream);

	TL vaLong		= static_cast<TL>(context->get<T>(ra));
	TL vbLong		= static_cast<TL>(context->get<T>(rb));

	TL resultLong	= vaLong * vbLong;

	T resultLo		= static_cast<T>(resultLong & ((static_cast<TL>(1) << (sizeof(T) * CHAR_BIT)) - 1));
	T resultHi		= static_cast<T>(resultLong >> (sizeof(T) * CHAR_BIT));

	context->set<T>(rdlo, resultLo);
	context->set<T>(rdhi, resultHi);

	return stream;
}


inline const uint8_t* callbackExternalCall(IntepreterContext* context, LTy retType, const uint8_t* stream) {

	unsigned int	reg				= 0;
	uintptr_t		callAddress		= 0;
	unsigned int	paramCount		= 0;
	uint64_t		signature		= 0;

	READ_REGISTER_INDEX(reg, stream);
	READ_IMMP(callAddress, stream);
	READ_IMMI8(paramCount, stream);
	READ_IMMI64(signature, stream);

	NanoUnusedVariable(paramCount); // should be removed
	NanoUnusedVariable(retType);
	NanoAssert((signature & 0x07) == (retType + 1));

	interpreter::doExternalCallFromSignature(context, callAddress, reg, signature);

	return stream;
}

const uint8_t* callbackGenFlags(IntepreterContext* context, const uint8_t* stream);


template<MemoryByteOrder byteOrder, typename TSource, typename TRegister>
inline const uint8_t* callbackStore(IntepreterContext* context, const uint8_t* stream) {
	NanoUnusedVariable(context);

	uint32_t rd, rb;
	int32_t offset;
	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rb, stream);
	READ_IMMI32(offset, stream);

	TSource		value		= static_cast<TSource>(context->get<TRegister>(rd));
	uintptr_t	base		= context->get<uintptr_t>(rb);

	if (byteOrder == MemoryByteOrder_Host) {
		interpreter::storeValue(base + offset, value);
	} else {
		interpreter::storeValueReversed(base + offset, value);
	}

	return stream;
}


template<MemoryByteOrder byteOrder, typename TSource, typename TRegister>
inline const uint8_t* callbackLoad(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rb; 
	int32_t offset;
	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rb, stream);
	READ_IMMI32(offset, stream);

	uintptr_t base = context->get<uintptr_t>(rb);
	
	TSource value;
	if (byteOrder == MemoryByteOrder_Host)
		value = interpreter::loadValue<TSource>(base + offset);
	else 
		value = interpreter::loadValueReversed<TSource>(base + offset);

	context->set<TRegister>(rd, static_cast<TRegister>(value));

	return stream;
}

inline const uint8_t* callbackCopy(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;

	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	context->copy(rd, rs);

	return stream;
}


inline const uint8_t* callbackGetFlg(IntepreterContext* context, const uint8_t* stream, ArithFlag a_eFlag) {

	uint32_t r;
	READ_REGISTER_INDEX(r, stream);
	context->set<uint32_t>(r, context->getFlag(a_eFlag) ? 1 : 0);

	return stream;
}

inline const uint8_t* callbackGetNotFlg(IntepreterContext* context, const uint8_t* stream, ArithFlag a_eFlag) {

	uint32_t r;
	READ_REGISTER_INDEX(r, stream);
	context->set<uint32_t>(r, context->getFlag(a_eFlag) ? 0 : 1);

	return stream;
}

inline const uint8_t* callbackSetFlg(IntepreterContext* context, const uint8_t* stream, ArithFlag a_eFlag) {

	uint32_t r;
	READ_REGISTER_INDEX(r, stream);

	uint32_t value = context->get<uint32_t>(r);

	if (value != 0)
		context->setFlag(a_eFlag);
	else
		context->clearFlag(a_eFlag);

	return stream;
}

inline const uint8_t* callbackSetNotFlg(IntepreterContext* context, const uint8_t* stream, ArithFlag a_eFlag) {

	uint32_t r;
	READ_REGISTER_INDEX(r, stream);

	uint32_t value = context->get<uint32_t>(r);

	if (value != 0)
		context->clearFlag(a_eFlag);
	else
		context->setFlag(a_eFlag);

	return stream;
}

template<typename T, typename TS>
inline const uint8_t* callbackCmov(IntepreterContext* context, const uint8_t* stream) {

	uint32_t cmpOp, cmpType;
	uint32_t rr, rca, rcb, rf, rt;

	READ_IMMI16(cmpOp, stream);
	READ_IMMI16(cmpType, stream);
	READ_REGISTER_INDEX(rr, stream);
	READ_REGISTER_INDEX(rca, stream);
	READ_REGISTER_INDEX(rcb, stream);
	READ_REGISTER_INDEX(rf, stream);
	READ_REGISTER_INDEX(rt, stream);

	bool conditionResult = doTest(context, rca, rcb, static_cast<nanojit::LOpcode>(cmpOp), static_cast<nanojit::LTy>(cmpType));

	T value = context->get<T>(conditionResult ? rt : rf);
	context->set<T>(rr, value);

	return stream;
}

inline const uint8_t* callbackBranch(const uint8_t* stream) {
	
 	ptrdiff_t offset = 0;
	READ_JUMP_OFFSET(offset, stream);
	return (stream + offset);
}


inline const uint8_t* callbackBranchCond(IntepreterContext* context, const uint8_t* stream) {

	ptrdiff_t offset = 0;
	const uint8_t* target = NULL;
	uint32_t cmpOp, cmpType;
	uint32_t rca, rcb;

	READ_JUMP_OFFSET(offset, stream);
	target = stream + offset;

	READ_IMMI16(cmpOp, stream);
	READ_IMMI16(cmpType, stream);
	READ_REGISTER_INDEX(rca, stream);
	READ_REGISTER_INDEX(rcb, stream);

	bool conditionResult = doTest(context, rca, rcb, static_cast<nanojit::LOpcode>(cmpOp), static_cast<nanojit::LTy>(cmpType));

	if (conditionResult)
		return target;
	
	return stream;
}


template<typename T>
inline const uint8_t* callbackSpill(IntepreterContext* context, const uint8_t* stream) {

	uint32_t d;
	uint32_t r;

	READ_REGISTER_INDEX(r, stream);
	READ_IMMI32(d, stream);

	T value = context->get<T>(r);
	uintptr_t iFrameAddress = context->get<uintptr_t>(nanojit::REGNUM(nanojit::FP));
	*reinterpret_cast<T*>(reinterpret_cast<uint8_t*>(iFrameAddress) + d) = value;

	return stream;
}

template<typename T>
inline const uint8_t* callbackRestore(IntepreterContext* context, const uint8_t* stream) {
	uint32_t d;
	uint32_t r;

	READ_REGISTER_INDEX(r, stream);
	READ_IMMI32(d, stream);

	uintptr_t iFrameAddress = context->get<uintptr_t>(nanojit::REGNUM(nanojit::FP));
	T value = *reinterpret_cast<T*>(reinterpret_cast<uint8_t*>(iFrameAddress) + d);
	context->set<T>(r, value);

	return stream;
}

template<typename T>
inline const uint8_t* callbackPush(IntepreterContext* context, const uint8_t* stream) {

	uint32_t r;
	unsigned int offset;

	READ_REGISTER_INDEX(r, stream);
	READ_IMMI32(offset, stream);

	T value = context->get<T>(r);
	context->setStack<T>(offset, value);

	return stream;
}

template<typename T>
inline const uint8_t* callbackPop(IntepreterContext* context, const uint8_t* stream) {

	uint32_t r;
	unsigned int offset;

	READ_REGISTER_INDEX(r, stream);
	READ_IMMI32(offset, stream);

	T value = context->getStack<T>(offset);
	context->set<T>(r, value);

	return stream;
}

template<typename TS, typename TD>
inline const uint8_t* callbackMov(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;
	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	context->clear(rd);

	if (sizeof(TS) == 4)
		context->set<uint32_t>(rd, context->get<uint32_t>(rs));
	else if (sizeof(TS) == 8)
		context->set<uint64_t>(rd, context->get<uint64_t>(rs));
	else
		NanoAssert(0);

	return stream;
}


template<>
inline const uint8_t* callbackMov<int32_t, int64_t>(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;
	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	int32_t srcValue = context->get<int32_t>(rs);
	context->set<int64_t>(rd, static_cast<int64_t>(srcValue));

	return stream;
}

template<typename TS, typename TD>
const uint8_t* callbackCvt(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;
	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	TS srcValue		= context->get<TS>(rs);
	TD dstValue		= convert<TS, TD, false>(srcValue);
	context->set<TD>(rd, dstValue);

	return stream;
}

template<typename TS, typename TD>
const uint8_t* callbackCvtu(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;
	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	TS srcValue		= context->get<TS>(rs);
	TD dstValue		= convert<TS, TD, false>(srcValue);
	context->set<TD>(rd, dstValue);

	return stream;
}

template<typename TS, typename TD>
const uint8_t* callbackCvtt(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;
	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	TS srcValue		= context->get<TS>(rs);
	TD dstValue		= convert<TS, TD, true>(srcValue);
	context->set<TD>(rd, dstValue);

	return stream;
}

template<typename TS, typename TD>
const uint8_t* callbackCvtut(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;
	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	TS srcValue		= context->get<TS>(rs);
	TD dstValue		= convert<TS, TD, true>(srcValue);
	context->set<TD>(rd, dstValue);

	return stream;
}

template<typename TD, typename TS>
const uint8_t* callbackCvf(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;
	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	TS srcValue		= context->get<TS>(rs);
	TD dstValue		= convert<TS, TD, false>(srcValue);
	context->set<TD>(rd, dstValue);

	return stream;
}


} // namespace interpreter
} // namespace nanojit

#endif // !defined(__NativeInterpreterVmCallbacks_h__)

