
#ifndef __nanojit_CodeBlock__
#define __nanojit_CodeBlock__

#include "nanojit.h"
#include <stddef.h>

namespace nanojit
{
	class HeapBlock;

	/**
	 * CodeBlock is a single block of code. The next field is used to
	 * form linked lists of non-contiguous blocks of code. Clients use CodeList*
	 * to point to the first block in a list.
	 */
	class CodeBlock
	{
		/** pointer to the heapblock terminal that represents the code chunk containing this block */
		HeapBlock*		_terminator;

		/** pointer to the end of the code block */
		NIns*			_end;

		/** for making a linked lists of blocks in any order */
		CodeBlock*		_next;

		/** lower code block inside the heap block */
		CodeBlock*		_lower;
		/** upper code block inside the heap block */
		CodeBlock*		_upper;


		/** offset of code content inside the block */
		/** sometimes the CodeAlloc can add padding */
		uint16_t		_codeStartOffsetInBytes;
		uint16_t		_codeEndOffsetInBytes;

		/** true if block is free, false otherwise */
		bool			_isFree;

#if defined(NANOJIT_MINIMUM_SPACING_BETWEEN_CODE_BLOCKS_IN_BYTES)
		static const size_t headerSize =
			sizeof(HeapBlock*)
			+	sizeof(NIns*)
			+	sizeof(CodeBlock*)
			+	sizeof(CodeBlock*)
			+	sizeof(CodeBlock*)
			+	sizeof(uint16_t)
			+	sizeof(uint16_t)
			+	sizeof(bool);

		static const size_t paddingSize = NANOJIT_MINIMUM_SPACING_BETWEEN_CODE_BLOCKS_IN_BYTES - headerSize;
		uint8_t padding[paddingSize];
#endif

		/** code holds this block's payload of binary code, from
			here to this->end */
		NIns			_code[1]; // more follows


		friend class CodeBlockList;

	public:
		static inline size_t sizeOfBlockHeader() { return (size_t)(&((CodeBlock*)0)->_code[0]); }

		void initialize(HeapBlock* terminator, NIns* end, CodeBlock* lower, CodeBlock* upper, CodeBlock* next);

		/** return the starting code address */
		inline NIns* internalBufferStart() { return &_code[0]; }

		/** return the starting code address - const overload */
		inline const NIns* internalBufferStart() const { return &_code[0]; }

		/** return the ending code address */
		inline NIns* internalBufferEnd() { return _end; }

		/** return the ending code address - const overload */
		inline const NIns* internalBufferEnd() const { return _end; }

		/** return the usable size of this block */
		inline size_t internalBufferSize() const { return (internalBufferEnd() - internalBufferStart()) * sizeof(NIns); }

		/** return the start of the code contained inside the block */
		inline const NIns* codeStart() const { return reinterpret_cast<const NIns*>(reinterpret_cast<const uint8_t*>(&_code[0]) + _codeStartOffsetInBytes); }

		/** retiurn a pointer to the opcode at the given offset  */
		inline NIns* accessCodeAt(ptrdiff_t offset) { NanoAssert(((offset % sizeof(NIns)) == 0) && (offset < static_cast<ptrdiff_t>(codeSize()))); return reinterpret_cast<NIns*>(reinterpret_cast<uint8_t*>(&_code[0]) + offset); }

		/** return the size of the code stored inside the glock */
		inline size_t codeSize() const { return internalBufferSize() - (_codeStartOffsetInBytes + _codeEndOffsetInBytes); }

		inline bool isFree() const { return _isFree; }
		inline void markFree() { _isFree = true; NanoAssert(_next == NULL); }
		inline void markUsed() { _isFree = false; }

		/** return the whole size of this block including overhead */
		inline size_t allocatedSize() const { return internalBufferSize() + sizeOfBlockHeader(); }

		/** true is the given NIns is contained within this block */
		bool isInBlock(const NIns* n) const { return (n >= internalBufferStart() && n < internalBufferEnd()); }

		/** return next block in list */
		inline CodeBlock* next() { return _next; }

		/** return next block in list - const overload */
		inline const CodeBlock* next() const { return _next; }

		/** return lower block in heap block - const overload */
		inline const CodeBlock* lower() const { return _lower; }
		inline CodeBlock* lower() { return _lower; }

		/** return upper block in heap block - const overload */
		inline const CodeBlock* upper() const { return _upper; }
		inline CodeBlock* upper() { return _upper; }

		/** return terminator heap block for this code block */
		inline HeapBlock* terminator() { return _terminator; }

		uintptr_t offsetFromCodeEndInBytes(const NIns* ins) const;
		uintptr_t offsetFromCodeStartInBytes(const NIns* ins) const;

		inline size_t codeStartOffset() const {
			return _codeStartOffsetInBytes;
		}

		inline void setCodeStartOffset(uintptr_t a_iDiff) {
			NanoAssert(a_iDiff < 0xFFFF);
			_codeStartOffsetInBytes = static_cast<uint16_t>(a_iDiff);
		}

		inline size_t codeEndOffset() const {
			return _codeEndOffsetInBytes;
		}

		inline void setCodeEndOffset(uintptr_t a_iDiff) {
			NanoAssert(a_iDiff < 0xFFFF);
			_codeEndOffsetInBytes = static_cast<uint16_t>(a_iDiff);
		}
		
		bool isBeforeInList(const CodeBlock* other) const;

		/** return the minimum free size to split a block, given a minimum allocation size */
		static size_t getMinimumFreeSizeForSplit(size_t minBlockSize);
		/** return true if the block can be splitted */
		static bool canSplitBlock(size_t minBlockSize, CodeBlock* block, NIns* splitPoint);
		/** split block at size and return new block - block instance is still valid */
		static CodeBlock* splitBlock(size_t minBlockSize, CodeBlock* block, NIns* splitPoint);
		/** merge codeBlock with the upper block in heap - update codeBlock instance */
		static void mergeCurrentWithUpper(CodeBlock* codeBlock);

	};

	inline uintptr_t CodeBlock::offsetFromCodeEndInBytes(const NIns* ins) const {
		NanoAssert(isInBlock(ins));
		return (uintptr_t(internalBufferEnd()) - _codeEndOffsetInBytes) - uintptr_t(ins);
	}

	inline uintptr_t CodeBlock::offsetFromCodeStartInBytes(const NIns* ins) const {
		NanoAssert(isInBlock(ins));
		return uintptr_t(ins) - uintptr_t(codeStart());
	}

	inline size_t CodeBlock::getMinimumFreeSizeForSplit(size_t minBlockSize) {
		return (minBlockSize * 2) + sizeOfBlockHeader() + alignof<CodeBlock>();
	}

	inline bool CodeBlock::canSplitBlock(size_t minBlockSize, CodeBlock* block, NIns* splitPoint) {
		// make sure the current block is big enough
		size_t minFreeSizeForSplit = getMinimumFreeSizeForSplit(minBlockSize);
		return (uintptr_t(splitPoint) - uintptr_t(block->internalBufferStart())) >= minFreeSizeForSplit;
	}

} // namespace nanojit

#endif 

